@@ -3,14 +3,13 @@
 #include "r_list.h"
 #include "r_config.h"
 #include "r_core.h"
-#include "r_print.h"
+#include "r_util.h"
 #include "r_bin.h"
 #include "r_debug.h"
 
 static const char *help_msg_o[] = {
 	"Usage: o","[com- ] [file] ([offset])","",
 	"o","","list opened files",
-	"o"," 4","Switch to open file on fd 4",
 	"o","-1","close file descriptor 1",
 	"o-","!*","close all opened files",
 	"o--","","close all files, analysis, binfiles, flags, same as !r2 --",
@@ -34,7 +33,7 @@ static const char *help_msg_o[] = {
 	"oo","+","reopen current file in read-write",
 	"ood","[r] [args]","reopen in debugger mode (with args)",
 	"oo[bnm]"," [...]","see oo? for help",
-	"op"," [fd]", "prioritize given fd (see also ob)",
+	"op"," [fd]", "select the given fd as current file (see also ob)",
 	"ox", " fd fdx", "exchange the descs of fd and fdx and keep the mapping",
 	NULL
 };
@@ -53,7 +52,7 @@ static const char *help_msg_o_star[] = {
 };
 
 static const char *help_msg_oa[] = {
-	"Usage:", "oa [addr] ([filename])", " # load bininfo and update flags",
+	"Usage:", "oba [addr] ([filename])", " # load bininfo and update flags",
 	"oba", " [addr]", "Open bin info from the given address",
 	"oba", " [addr] [filename]", "Open file and load bin info at given address",
 	NULL
@@ -63,16 +62,18 @@ static const char *help_msg_ob[] = {
 	"Usage:", "ob", " # List open binary files backed by fd",
 	"ob", "", "List opened binary files and objid",
 	"ob*", "", "List opened binary files and objid (r2 commands)",
+// those 3 commands are VERY SIMILAR, need love
 	"ob", " [fd objid]", "Switch to open binary file by fd number and objid",
+	"obo", " [objid]", "Switch to open binary file by objid",
+	"obb", " [fd]", "Switch to open binfile by fd number",
+
 	"oba", " [addr]", "Open bin info from the given address",
 	"oba", " [addr] [filename]", "Open file and load bin info at given address",
-	"obb", " [fd]", "Switch to open binfile by fd number",
 	"obf", " ([file])", "Load bininfo for current file (useful for r2 -n)",
 	"obj", "", "List opened binary files and objid (JSON format)",
 	"obr", " [baddr]", "Rebase current bin object",
 	"ob-", "[objid]", "Delete binfile by binobjid",
 	"ob-", "*", "Delete all binfiles",
-	"obo", " [objid]", "Switch to open binary file by objid",
 	NULL
 };
 
@@ -89,18 +90,22 @@ static const char *help_msg_om[] = {
 	"om=", "", "list all maps in ascii art",
 	"omj", "", "list all maps in json format",
 	"om", " [fd]", "list all defined IO maps for a specific fd",
-	"om", "-mapid", "remove the map with corresponding id",
+	"om-", "mapid", "remove the map with corresponding id",
+	"om-..", "", "hud view of all the maps to select the one to delete",
 	"om", " fd vaddr [size] [paddr] [rwx] [name]", "create new io map",
 	"omm"," [fd]", "create default map for given fd. (omm `oq`)",
 	"om.", "", "show map, that is mapped to current offset",
-	"omn", " mapid [name]", "set/delete name for map with mapid",
+	"omn", " mapaddr [name]", "set/delete name for map which spans mapaddr",
+	"omni", " mapid [name]", "set/delete name for map with mapid",
 	"omn.", "([-|name])", "show/set/delete name for current map",
 	"omf", " [mapid] rwx", "change flags/perms for current/given map",
 	"omfg", "[+-]rwx", "change flags/perms for all maps (global)",
 	"omb", " mapid addr", "relocate map with corresponding id",
 	"omb.", " addr", "relocate current map",
 	"omr", " mapid newsize", "resize map with corresponding id",
+	"omo", " fd", "map the given fd with lowest priority",
 	"omp", " mapid", "prioritize map with corresponding id",
+	"ompd", " mapid", "deprioritize map with corresponding id",
 	"ompf", "[fd]", "prioritize map by fd",
 	"ompb", " binid", "prioritize maps of mapped bin with binid",
 	"omps", " sectionid", "prioritize maps of mapped section with sectionid",
@@ -122,43 +127,32 @@ static const char *help_msg_oo[] = {
 };
 
 static const char *help_msg_oo_plus[] = {
-	"oo+", "", "reopen in read-write",
+	"Usage:", "oo+", " # reopen in read-write",
 	NULL
 };
 
 static const char *help_msg_oob[] = {
-	"oob", "", "reopen loading rbin info",
+	"Usage:", "oob", " # reopen loading rbin info",
 	NULL
 };
 
 static const char *help_msg_ood[] = {
-	"ood"," [args]","reopen in debugger mode (with args)",
+	"Usage:", "ood [args]", " # reopen in debugger mode (with args)",
 	"oodr"," [rarun2]","same as dor ..;ood",
 	NULL
 };
 
 static const char *help_msg_oon[] = {
-	"oon", "", "reopen without loading rbin info",
+	"Usage:", "oon", " # reopen without loading rbin info",
 	NULL
 };
 
 static const char *help_msg_oonn[] = {
-	"oonn", "", "reopen without loading rbin info, but with header flags",
+	"Usage:", "oonn", " # reopen without loading rbin info, but with header flags",
 	NULL
 };
 
-static inline ut32 find_binfile_id_by_fd (RBin *bin, ut32 fd) {
-	RListIter *it;
-	RBinFile *bf;
-	r_list_foreach (bin->binfiles, it, bf) {
-		if (bf->fd == fd) {
-			return bf->id;
-		}
-	}
-	return UT32_MAX;
-}
-
-static RBinObject *find_binfile_by_id (RBin *bin, ut32 id) {
+static RBinObject *find_binfile_by_id(RBin *bin, ut32 id) {
 	RListIter *it, *it2;
 	RBinFile *bf;
 	RBinObject *obj;
@@ -212,9 +206,9 @@ static void list_maps_visual(RIO *io, ut64 seek, ut64 len, int width, int use_co
 		ls_foreach_prev (io->maps, iter, s) {
 			if (use_color) {
 				color_end = Color_RESET;
-				if (s->flags & 1) { // exec bit
+				if (s->perm & R_PERM_X) { // exec bit
 					color = Color_GREEN;
-				} else if (s->flags & 2) { // write bit
+				} else if (s->perm & R_PERM_W) { // write bit
 					color = Color_RED;
 				} else {
 					color = "";
@@ -241,7 +235,7 @@ static void list_maps_visual(RIO *io, ut64 seek, ut64 len, int width, int use_co
 			}
 			io->cb_printf ("| %s0x%08"PFMT64x"%s %s %d %s\n",
 					color, r_itv_end (s->itv), color_end,
-					r_str_rwx_i (s->flags), s->fd, s->name);
+					r_str_rwx_i (s->perm), s->fd, s->name);
 			i++;
 		}
 		/* current seek */
@@ -282,13 +276,17 @@ static void cmd_open_bin(RCore *core, const char *input) {
 			char *arg = strdup (input + 3);
 			char *filename = strchr (arg, ' ');
 			if (filename) {
-				RIODesc *desc = r_io_open (core->io, filename + 1, R_IO_READ, 0);
+				int saved_fd = r_io_fd_get_current (core->io);
+				RIODesc *desc = r_io_open (core->io, filename + 1, R_PERM_R, 0);
 				if (desc) {
 					*filename = 0;
 					ut64 addr = r_num_math (core->num, arg);
-					r_bin_load_io (core->bin, desc->fd, addr, 0, 0);
+					RBinOptions opt;
+					r_bin_options_init (&opt, desc->fd, addr, 0, core->bin->rawstr);
+					r_bin_open_io (core->bin, &opt);
 					r_io_desc_close (desc);
 					r_core_cmd0 (core, ".is*");
+					r_io_use_fd (core->io, saved_fd);
 				} else {
 					eprintf ("Cannot open %s\n", filename + 1);
 				}
@@ -297,7 +295,9 @@ static void cmd_open_bin(RCore *core, const char *input) {
 				int fd = r_io_fd_get_current (core->io);
 				RIODesc *desc = r_io_desc_get (core->io, fd);
 				if (desc) {
-					r_bin_load_io (core->bin, desc->fd, addr, 0, 0);
+					RBinOptions opt;
+					r_bin_options_init (&opt, desc->fd, addr, 0, core->bin->rawstr);
+					r_bin_open_io (core->bin, &opt);
 					r_core_cmd0 (core, ".is*");
 				} else {
 					eprintf ("No file to load bin from?\n");
@@ -310,33 +310,35 @@ static void cmd_open_bin(RCore *core, const char *input) {
 			RIODesc *desc;
 			RListIter *iter;
 			r_list_foreach (files, iter, desc) {
-				r_bin_load_io (core->bin, desc->fd, core->offset, 0, 0);
+				RBinOptions opt;
+				r_bin_options_init (&opt, desc->fd, core->offset, 0, core->bin->rawstr);
+				r_bin_open_io (core->bin, &opt);
 				r_core_cmd0 (core, ".is*");
 				break;
 			}
 			r_list_free (files);
 		}
-		//r_bin_load_io_at_offset_as (core->bin, core->file->desc,
 		break;
 	case 'b': // "obb"
 		{
 			ut32 fd;
-			value = *(input + 3) ? input + 3 : NULL;
+			if (input[2] && input[3]) {
+				value = *(input + 3) ? input + 3 : NULL;
+			} else {
+				value = NULL;
+			}
 			if (!value) {
 				eprintf ("Invalid fd number.");
 				break;
 			}
-			binfile_num = UT32_MAX;
-			fd = *value && r_is_valid_input_num_value (core->num, value) ?
-				r_get_input_num_value (core->num, value) : UT32_MAX;
-			binfile_num = find_binfile_id_by_fd (core->bin, fd);
-			if (binfile_num == UT32_MAX) {
+			fd = *value && r_is_valid_input_num_value (core->num, value)? r_get_input_num_value (core->num, value): UT32_MAX;
+			RBinFile *bf = r_bin_file_find_by_fd (core->bin, fd);
+			if (!bf) {
 				eprintf ("Invalid fd number.");
 				break;
 			}
-			r_core_bin_raise (core, binfile_num, -1);
-		}
-		break;
+			r_core_bin_raise (core, bf->id, -1);
+	} break;
 	case ' ': // "ob "
 	{
 		ut32 fd;
@@ -356,14 +358,13 @@ static void cmd_open_bin(RCore *core, const char *input) {
 			break;
 		}
 		tmp = r_str_word_get0 (v, 0);
-		fd  = *v && r_is_valid_input_num_value (core->num, tmp) ?
-			r_get_input_num_value (core->num, tmp) : UT32_MAX;
+		fd = *v && r_is_valid_input_num_value (core->num, tmp)? r_get_input_num_value (core->num, tmp): UT32_MAX;
 		if (n == 2) {
 			tmp = r_str_word_get0 (v, 1);
-			binobj_num  = *v && r_is_valid_input_num_value (core->num, tmp) ?
-				r_get_input_num_value (core->num, tmp) : UT32_MAX;
+			binobj_num = *v && r_is_valid_input_num_value (core->num, tmp)? r_get_input_num_value (core->num, tmp): UT32_MAX;
 		} else {
-			binfile_num = find_binfile_id_by_fd (core->bin, fd);
+			RBinFile *bf = r_bin_file_find_by_fd (core->bin, fd);
+			binfile_num = bf? bf->id: UT32_MAX;
 		}
 		r_core_bin_raise (core, binfile_num, binobj_num);
 		free (v);
@@ -374,10 +375,12 @@ static void cmd_open_bin(RCore *core, const char *input) {
 		r_core_cmd0 (core, ".is*");
 		break;
 	case 'f':
-		// TODO: specify path to file?
-		r_core_bin_load (core, NULL, UT64_MAX);
-		value = input[2] ? input + 2 : NULL;
-		// r2_obf (core, value);
+		if (input[2] == ' ') {
+			r_core_cmdf (core, "oba 0 %s", input + 3);
+		} else {
+			r_core_bin_load (core, NULL, UT64_MAX);
+			value = input[2] ? input + 2 : NULL;
+		}
 		break;
 	case 'o': // "obo"
 		value = input[2] ? input + 2 : NULL;
@@ -449,22 +452,22 @@ static void map_list(RIO *io, int mode, RPrint *print, int fd) {
 			}
 			first = false;
 			print->cb_printf ("{\"map\":%i,\"fd\":%d,\"delta\":%"PFMT64u",\"from\":%"PFMT64u
-					",\"to\":%"PFMT64u",\"flags\":\"%s\",\"name\":\"%s\"}", map->id, map->fd,
+					",\"to\":%"PFMT64u",\"perm\":\"%s\",\"name\":\"%s\"}", map->id, map->fd,
 					map->delta, map->itv.addr, r_itv_end (map->itv),
-					r_str_rwx_i (map->flags), (map->name ? map->name : ""));
+					r_str_rwx_i (map->perm), (map->name ? map->name : ""));
 			break;
 		case 1:
 		case '*':
 		case 'r':
-			print->cb_printf ("om %d 0x%"PFMT64x" 0x%"PFMT64x" 0x%"PFMT64x" %s%s%s\n", map->fd,
-					map->itv.addr, map->itv.size, map->delta, r_str_rwx_i(map->flags),
+			print->cb_printf ("om %d 0x%08"PFMT64x" 0x%08"PFMT64x" 0x%08"PFMT64x" %s%s%s\n", map->fd,
+					map->itv.addr, map->itv.size, map->delta, r_str_rwx_i(map->perm),
 					map->name ? " " : "", map->name ? map->name : "");
 			break;
 		default:
 			print->cb_printf ("%2d fd: %i +0x%08"PFMT64x" 0x%08"PFMT64x
 					" - 0x%08"PFMT64x" %s %s\n", map->id, map->fd,
 					map->delta, map->itv.addr, r_itv_end (map->itv) - 1,
-					r_str_rwx_i (map->flags), (map->name ? map->name : ""));
+					r_str_rwx_i (map->perm), (map->name ? map->name : ""));
 			break;
 		}
 	}
@@ -473,12 +476,12 @@ static void map_list(RIO *io, int mode, RPrint *print, int fd) {
 	}
 }
 
-static void cmd_omfg (RCore *core, const char *input) {
+static void cmd_omfg(RCore *core, const char *input) {
 	SdbListIter *iter;
 	RIOMap *map;
 	input = r_str_trim_ro (input);
 	if (input) {
-		int flags = *input
+		int perm = *input
 		? (*input == '+' || *input == '-')
 			? r_str_rwx (input + 1)
 			: r_str_rwx (input)
@@ -486,24 +489,24 @@ static void cmd_omfg (RCore *core, const char *input) {
 		switch (*input) {
 		case '+':
 			ls_foreach (core->io->maps, iter, map) {
-				map->flags |= flags;
+				map->perm |= perm;
 			}
 			break;
 		case '-':
 			ls_foreach (core->io->maps, iter, map) {
-				map->flags &= ~flags;
+				map->perm &= ~perm;
 			}
 			break;
 		default:
 			ls_foreach (core->io->maps, iter, map) {
-				map->flags = flags;
+				map->perm = perm;
 			}
 			break;
 		}
 	}
 }
 
-static void cmd_omf (RCore *core, const char *input) {
+static void cmd_omf(RCore *core, const char *input) {
 	SdbListIter *iter;
 	RIOMap *map;
 	char *arg = strdup (r_str_trim_ro (input));
@@ -515,19 +518,19 @@ static void cmd_omf (RCore *core, const char *input) {
 		// change perms of Nth map
 		*sp++ = 0;
 		int id = r_num_math (core->num, arg);
-		int flags = (*sp)? r_str_rwx (sp): 7;
+		int perm = (*sp)? r_str_rwx (sp): R_PERM_RWX;
 		ls_foreach (core->io->maps, iter, map) {
 			if (map->id == id) {
-				map->flags = flags;
+				map->perm = perm;
 				break;
 			}
 		}
 	} else {
 		// change perms of current map
-		int flags = (arg && *arg)? r_str_rwx (arg): 7;
+		int perm = (arg && *arg)? r_str_rwx (arg): R_PERM_RWX;
 		ls_foreach (core->io->maps, iter, map) {
 			if (r_itv_contain (map->itv, core->offset)) {
-				map->flags = flags;
+				map->perm = perm;
 			}
 		}
 	}
@@ -537,6 +540,7 @@ static void cmd_omf (RCore *core, const char *input) {
 static void cmd_open_map(RCore *core, const char *input) {
 	ut64 fd = 0LL;
 	ut32 id = 0;
+	ut64 addr = 0;
 	char *s = NULL, *p = NULL, *q = NULL;
 	ut64 new;
 	RIOMap *map = NULL;
@@ -549,7 +553,7 @@ static void cmd_open_map(RCore *core, const char *input) {
 			core->print->cb_printf ("map: %i fd: %i +0x%"PFMT64x" 0x%"PFMT64x
 				" - 0x%"PFMT64x" ; %s : %s\n", map->id, map->fd,
 				map->delta, map->itv.addr, r_itv_end (map->itv),
-			r_str_rwx_i (map->flags), map->name ? map->name : "");
+			r_str_rwx_i (map->perm), map->name ? map->name : "");
 		}
 		break;
 	case 'r': // "omr"
@@ -582,8 +586,24 @@ static void cmd_open_map(RCore *core, const char *input) {
 			}
 		}
 		break;
+	case 'o': // "omo"
+		if (input[2] == ' ') {
+			r_core_cmdf (core, "om %s 0x%08" PFMT64x " $s r omo", input + 2);
+		} else {
+			r_core_cmd0 (core, "om `oq.` $B $s r");
+		}
+		r_core_cmd0 (core, "ompd `omq.`");
+		break;
 	case 'p':
 		switch (input[2]) {
+		case 'd': // "ompf"
+			id = r_num_math (core->num, input + 3);		//mapid
+			if (r_io_map_exists_for_id (core->io, id)) {
+				r_io_map_depriorize (core->io, id);
+			} else {
+				eprintf ("Cannot find any map with mapid %d\n", id);
+			}
+			break;
 		case 'f': // "ompf"
 			fd = r_num_math (core->num, input + 3);
 			if (!r_io_map_priorize_for_fd (core->io, (int)fd)) {
@@ -606,6 +626,7 @@ static void cmd_open_map(RCore *core, const char *input) {
 			id = r_num_math (core->num, input + 3);		//mapid
 			if (r_io_map_exists_for_id (core->io, id)) {
 				r_io_map_priorize (core->io, id);
+				r_core_block_read (core);
 			} else {
 				eprintf ("Cannot find any map with mapid %d\n", id);
 			}
@@ -648,7 +669,7 @@ static void cmd_open_map(RCore *core, const char *input) {
 				if (!size) {
 					size = r_io_fd_size (core->io, fd);
 				}
-				map = r_io_map_add (core->io, fd, rwx_arg ? rwx : desc->flags, paddr, vaddr, size, true);
+				map = r_io_map_add (core->io, fd, rwx_arg ? rwx : desc->perm, paddr, vaddr, size);
 				r_io_map_set_name (map, name);
 			}
 		} else {
@@ -673,15 +694,18 @@ static void cmd_open_map(RCore *core, const char *input) {
 					r_cons_printf ("%s\n", map->name);
 					break;
 				default:
-					r_io_map_set_name (map, input + 3);
+					r_io_map_set_name (map, r_str_trim_ro (input + 3));
 					break;
 				}
 			}
 		} else {
-			if (!(s = strdup (&input[2]))) {
+			bool use_id = (input[2] == 'i') ? true : false;
+			s = strdup ( use_id ? &input[3] : &input[2]);
+			if (!s) {
 				break;
 			}
 			p = s;
+
 			while (*s == ' ') {
 				s++;
 			}
@@ -690,16 +714,26 @@ static void cmd_open_map(RCore *core, const char *input) {
 				break;
 			}
 			if (!(q = strchr (s, ' '))) {
-				id = (ut32)r_num_math (core->num, s);
-				map = r_io_map_get (core->io, id);
+				if (use_id) {
+					id = (ut32)r_num_math (core->num, s);
+					map = r_io_map_resolve (core->io, id);
+				} else {
+					addr = r_num_math (core->num, s);
+					map = r_io_map_get (core->io, addr);
+				}
 				r_io_map_del_name (map);
 				s = p;
 				break;
 			}
 			*q = '\0';
 			q++;
-			id = (ut32)r_num_math (core->num, s);
-			map = r_io_map_get (core->io, id);
+			if (use_id) {
+				id = (ut32)r_num_math (core->num, s);
+				map = r_io_map_resolve (core->io, id);
+			} else {
+				addr = r_num_math (core->num, s);
+				map = r_io_map_get (core->io, addr);
+			}
 			if (*q) {
 				r_io_map_set_name (map, q);
 			} else {
@@ -710,15 +744,11 @@ static void cmd_open_map(RCore *core, const char *input) {
 		break;
 	case 'm': // "omm"
 		{
-			ut64 fd = input[3]? r_num_math (core->num, input + 3): UT64_MAX;
+			ut32 fd = input[2]? r_num_math (core->num, input + 2): r_io_fd_get_current (core->io);
 			RIODesc *desc = r_io_desc_get (core->io, fd);
-			if (!desc) {
-				fd = r_io_fd_get_current (core->io);
-				desc = r_io_desc_get (core->io, fd);
-			}
 			if (desc) {
 				ut64 size = r_io_desc_size (desc);
-				map = r_io_map_add (core->io, fd, desc->flags, 0, 0, size, true);
+				map = r_io_map_add (core->io, fd, desc->perm, 0, 0, size);
 				r_io_map_set_name (map, desc->name);
 			} else {
 				eprintf ("Usage: omm [fd]\n");
@@ -726,7 +756,9 @@ static void cmd_open_map(RCore *core, const char *input) {
 		}
 		break;
 	case '-': // "om-"
-		if (input[2] == '*') {
+		if (!strcmp (input + 2, "..")) {
+			r_core_cmd0 (core, "om-`om~...`~[0]");
+		} else if (input[2] == '*') {
 			r_io_map_reset (core->io);
 		} else {
 			r_io_map_del (core->io, r_num_math (core->num, input + 2));
@@ -749,7 +781,14 @@ static void cmd_open_map(RCore *core, const char *input) {
 	case 'j': // "omj"
 	case '*': // "om*"
 	case 'q': // "omq"
-		map_list (core->io, input[1], core->print, -1);
+		if (input[1] && input[2] == '.') {
+			map = r_io_map_get (core->io, core->offset);
+			if (map) {
+				core->print->cb_printf ("%i\n", map->id);
+			}
+		} else {
+			map_list (core->io, input[1], core->print, -1);
+		}
 		break;
 	case '=': // "om="
 		list_maps_visual (core->io, core->offset, core->blocksize,
@@ -764,37 +803,55 @@ static void cmd_open_map(RCore *core, const char *input) {
 	r_core_block_read (core);
 }
 
-R_API void r_core_file_reopen_in_malloc (RCore *core) {
-	RListIter *iter;
-	RList *files = r_id_storage_list (core->io->files);
-	RIODesc *desc;
-	r_list_foreach (files, iter, desc) {
-		if (strstr (desc->name, "://")) {
-			continue;
-		}
-		ut64 sz = r_io_desc_size (desc);
-		ut8 *buf = calloc (sz, 1);
-		if (!buf) {
-			eprintf ("Cannot allocate %d\n", (int)sz);
-			continue;
-		}
-		(void)r_io_pread_at (core->io, 0, buf, sz);
-		char *url = r_str_newf ("malloc://%d", (int)sz);
-		// use r_io_desc_exchange pls
-		RIODesc *newDesc = r_io_open (core->io, url, R_IO_READ | R_IO_WRITE, 0);
-		if (newDesc) {
-			r_io_desc_close (desc);
-			(void)r_io_write_at (core->io, 0, buf, sz);
-		} else {
-			eprintf ("Cannot open %s\n", url);
-		}
+static bool reopen_in_malloc_cb(void *user, void *data, ut32 id) {
+	RIO *io = (RIO *)user;
+	RIODesc *desc = (RIODesc *)data;
+
+	if (r_io_desc_is_blockdevice (desc) || r_io_desc_is_dbg (desc)) {
+		return true;
+	}
+
+	if (strstr (desc->uri, "://")) {
+		return true;
+	}
+
+	ut64 size = r_io_desc_size (desc);
+
+	char *uri = r_str_newf ("malloc://%"PFMT64u, size);
+	if (!uri) {
+		return false;
+	}
+
+	ut8 *buf = malloc (size);
+// if malloc fails, we can just abort the loop by returning false
+	if (!buf) {
+		free (uri);
+		return false;
+	}
+
+	RIODesc *ndesc = r_io_open_nomap (io, uri, R_PERM_RW, 0);
+	free (uri);
+	if (!ndesc) {
 		free (buf);
-		free (url);
-		break;
+		return false;
+	}
+
+	r_io_desc_read_at (desc, 0LL, buf, (int)size);	//that cast o_O
+	r_io_desc_write_at (ndesc, 0LL, buf, (int)size);
+	free (buf);
+	r_io_desc_exchange (io, desc->fd, ndesc->fd);
+
+	r_io_desc_close (desc);
+	return true;
+}
+
+R_API void r_core_file_reopen_in_malloc(RCore *core) {
+	if (core && core->io && core->io->files) {
+		r_id_storage_foreach (core->io->files, reopen_in_malloc_cb, core->io);
 	}
 }
 
-R_API void r_core_file_reopen_debug (RCore *core, const char *args) {
+R_API void r_core_file_reopen_debug(RCore *core, const char *args) {
 	RCoreFile *ofile = core->file;
 	RBinFile *bf = NULL;
 	RIODesc *desc;
@@ -804,7 +861,7 @@ R_API void r_core_file_reopen_debug (RCore *core, const char *args) {
 		return;
 	}
 	bf = r_bin_file_find_by_fd (core->bin, ofile->fd);
-	binpath = bf ? strdup (bf->file) : NULL;
+	binpath = (bf && bf->file) ? strdup (bf->file) : NULL;
 	if (!binpath) {
 		if (r_file_exists (desc->name)) {
 			binpath = strdup (desc->name);
@@ -816,8 +873,9 @@ R_API void r_core_file_reopen_debug (RCore *core, const char *args) {
 		return;
 	}
 	int bits = core->assembler->bits;
-	char *oldname = r_file_abspath (binpath);
-	char *newfile = r_str_newf ("dbg://%s %s", oldname, args);
+	char *bin_abspath = r_file_abspath (binpath);
+	char *escaped_path = r_str_arg_escape (bin_abspath);
+	char *newfile = r_str_newf ("dbg://%s %s", escaped_path, args);
 	char *newfile2 = strdup (newfile);
 	desc->uri = newfile;
 	desc->referer = NULL;
@@ -841,7 +899,8 @@ R_API void r_core_file_reopen_debug (RCore *core, const char *args) {
 	}
 #endif
 	r_core_cmd0 (core, "sr PC");
-	free (oldname);
+	free (bin_abspath);
+	free (escaped_path);
 	free (binpath);
 	free (newfile);
 }
@@ -862,7 +921,7 @@ static bool desc_list_visual_cb(void *user, void *data, ut32 id) {
 	RIODesc *desc = (RIODesc *)data;
 	ut64 sz = r_io_desc_size (desc);
 	r_cons_printf ("%2d %c %s 0x%08"PFMT64x" ", desc->fd,
-			(desc->io && (desc->io->desc == desc)) ? '*' : '-', r_str_rwx_i (desc->flags), sz);
+			(desc->io && (desc->io->desc == desc)) ? '*' : '-', r_str_rwx_i (desc->perm), sz);
 	int flags = p->flags;
 	p->flags &= ~R_PRINT_FLAGS_HEADER;
 	r_print_progressbar (p, sz * 100 / fdsz, r_cons_get_size (NULL) - 40);
@@ -885,6 +944,13 @@ static bool desc_list_visual_cb(void *user, void *data, ut32 id) {
 	return true;
 }
 
+static bool desc_list_quiet2_cb(void *user, void *data, ut32 id) {
+	RPrint *p = (RPrint *)user;
+	RIODesc *desc = (RIODesc *)data;
+	p->cb_printf ("%d\n", desc->fd);
+	return false;
+}
+
 static bool desc_list_quiet_cb(void *user, void *data, ut32 id) {
 	RPrint *p = (RPrint *)user;
 	RIODesc *desc = (RIODesc *)data;
@@ -897,18 +963,32 @@ static bool desc_list_cb(void *user, void *data, ut32 id) {
 	RIODesc *desc = (RIODesc *)data;
 	p->cb_printf ("%2d %c %s 0x%08"PFMT64x" %s\n", desc->fd,
 			(desc->io && (desc->io->desc == desc)) ? '*' : '-',
-			r_str_rwx_i (desc->flags), r_io_desc_size (desc), desc->uri);
+			r_str_rwx_i (desc->perm), r_io_desc_size (desc), desc->uri);
+	return true;
+}
+
+static bool desc_list_json_cb(void *user, void *data, ut32 id) {
+	RPrint *p = (RPrint *)user;
+	RIODesc *desc = (RIODesc *)data;
+	// TODO: from is always 0? See libr/core/file.c:945
+	ut64 from = 0LL;
+	p->cb_printf ("{\"raised\":%s,\"fd\":%d,\"uri\":\"%s\",\"from\":%"
+			PFMT64d ",\"writable\":%s,\"size\":%" PFMT64d "}%s",
+			(desc->io && (desc->io->desc == desc)) ? "true" : "false",
+			desc->fd, desc->uri, from,
+			((desc->perm & R_PERM_W) ? "true": "false"),
+			r_io_desc_size (desc), (desc->io->files->top_id == id) ? "" : ",");
 	return true;
 }
 
 static int cmd_open(void *data, const char *input) {
 	RCore *core = (RCore*)data;
-	int perms = R_IO_READ;
-	ut64 baddr = r_config_get_i (core->config, "bin.baddr"),
-	     addr = 0LL;
-	int nowarn = r_config_get_i (core->config, "file.nowarn"),
-	    argc, fd;
+	int perms = R_PERM_R;
+	ut64 baddr = r_config_get_i (core->config, "bin.baddr");
+	ut64 addr = 0LL;
+	int argc, fd = -1;
 	RCoreFile *file;
+	RIODesc *desc;
 	bool silence = false;
 	bool write = false;
 	const char *ptr = NULL;
@@ -973,7 +1053,7 @@ static int cmd_open(void *data, const char *input) {
 			return 0;
 		} else if (input[1] == '+') { // "on+"
 			write = true;
-			perms |= R_IO_WRITE;
+			perms |= R_PERM_W;
 			if (input[2] != ' ') {
 				eprintf ("Usage: on+ file [addr] [rwx]\n");
 				return 0;
@@ -983,7 +1063,7 @@ static int cmd_open(void *data, const char *input) {
 			silence = true;
 			if (input[2] == '+') { // "ons+"
 				write = true;
-				perms |= R_IO_WRITE;
+				perms |= R_PERM_W;
 				if (input[3] != ' ') {
 					eprintf ("Usage: ons+ file [addr] [rwx]\n");
 					return 0;
@@ -1026,10 +1106,8 @@ static int cmd_open(void *data, const char *input) {
 			addr = r_num_math (core->num, argv[1]);
 			perms = r_str_rwx (argv[2]);
 		}
-		if ((file = r_core_file_open (core, ptr, perms, addr))) {
-			fd = file->fd;
-			r_io_map_add (core->io, fd, perms, 0LL, addr,
-					r_io_fd_size (core->io, fd), true);
+		if ((desc = r_io_open_at (core->io, ptr, perms, 0644, addr))) {
+			fd = desc->fd;
 		}
 		r_str_argv_free (argv);
 		if (!silence) {
@@ -1077,7 +1155,7 @@ static int cmd_open(void *data, const char *input) {
 					eprintf ("%d\n", fd);
 				}
 				r_core_bin_load (core, argv0, baddr);
-			} else if (!nowarn) {
+			} else {
 				eprintf ("cannot open file %s\n", argv0);
 			}
 			r_str_argv_free (argv);
@@ -1088,23 +1166,25 @@ static int cmd_open(void *data, const char *input) {
 #endif
 	case 'p': // "op"
 		/* handle prioritize */
-		{
+		if (input[1]) {
 			int fd = r_num_math (core->num, input + 1);
 			if (fd) {
 				RIODesc *desc = r_io_desc_get (core->io, fd);
 				if (desc) {
 					// only useful for io.va=0
-					core->io->desc = desc;
 					// load bininfo for given fd
-					r_core_cmdf (core, "ob %d", fd);
+					r_core_cmdf (core, "obb %d", fd);
+					core->io->desc = desc;
 				}
 			}
 			r_core_block_read (core);
+		} else {
+			r_cons_printf ("%d\n", core->io->desc->fd);
 		}
 		return 0;
 		break;
 	case '+': // "o+"
-		perms |= R_IO_WRITE;
+		perms |= R_PERM_W;
 	case 's': // "os"
 		silence = true;
 	case ' ': // "o" "o "
@@ -1144,7 +1224,7 @@ static int cmd_open(void *data, const char *input) {
 					eprintf ("%d\n", fd);
 				}
 				r_core_bin_load (core, argv0, baddr);
-			} else if (!nowarn) {
+			} else {
 				eprintf ("cannot open file %s\n", argv0);
 			}
 			r_str_argv_free (argv);
@@ -1160,7 +1240,11 @@ static int cmd_open(void *data, const char *input) {
 		r_id_storage_foreach (core->io->files, desc_list_visual_cb, core->print);
 		break;
 	case 'q': // "oq"
-		r_id_storage_foreach (core->io->files, desc_list_quiet_cb, core->print);
+		if (input[1] == '.') {
+			r_id_storage_foreach (core->io->files, desc_list_quiet2_cb, core->print);
+		} else {
+			r_id_storage_foreach (core->io->files, desc_list_quiet_cb, core->print);
+		}
 		break;
 	case '\0': // "o"
 		r_id_storage_foreach (core->io->files, desc_list_cb, core->print);
@@ -1177,7 +1261,9 @@ static int cmd_open(void *data, const char *input) {
 			r_core_cmd_help (core, help_msg_oj);
 			break;
 		}
-		r_core_file_list (core, (int)(*input));
+		core->print->cb_printf("[");
+		r_id_storage_foreach (core->io->files, desc_list_json_cb, core->print);
+		core->print->cb_printf("]\n");
 		break;
 	case 'L': // "oL"
 		if (r_sandbox_enable (0)) {
@@ -1194,7 +1280,6 @@ static int cmd_open(void *data, const char *input) {
 			} else {
 				r_io_plugin_list (core->io);
 			}
-			eprintf ("Usage: op [r2plugin."R_LIB_EXT"]\n");
 		}
 		break;
 	case 'i': // "oi"
@@ -1236,25 +1321,25 @@ static int cmd_open(void *data, const char *input) {
 			break;
 		}
 		break;
-	case 'u':
-		{
-			RListIter *iter = NULL;
-			RCoreFile *f;
-			int binfile_num;
-			core->switch_file_view = 0;
-			int num = atoi (input + 2);
+	case 'u': {
+		RListIter *iter = NULL;
+		RCoreFile *f;
+		int binfile_num;
+		core->switch_file_view = 0;
+		int num = atoi (input + 2);
 
-			r_list_foreach (core->files, iter, f) {
-				if (f->fd == num) {
-					core->file = f;
-				}
+		r_list_foreach (core->files, iter, f) {
+			if (f->fd == num) {
+				core->file = f;
 			}
-			r_io_use_fd (core->io, num);
-			binfile_num = find_binfile_id_by_fd (core->bin, num);
-			r_core_bin_raise (core, binfile_num, -1);
 		}
+		r_io_use_fd (core->io, num);
+		RBinFile *bf = r_bin_file_find_by_fd (core->bin, num);
+		binfile_num = bf? bf->id: UT32_MAX;
+		r_core_bin_raise (core, binfile_num, -1);
 		r_core_block_read (core);
 		break;
+	}
 	case 'b': // "ob"
 		cmd_open_bin (core, input);
 		break;
@@ -1267,7 +1352,9 @@ static int cmd_open(void *data, const char *input) {
 			r_core_file_close_fd (core, -1);
 			r_io_close_all (core->io);
 			r_bin_file_delete_all (core->bin);
-			r_list_purge (core->files);
+			if (core->files) {
+				r_list_purge (core->files);
+			}
 			break;
 		case '-': // "o--"
 			eprintf ("All core files, io, anal and flags info purged.\n");
@@ -1292,7 +1379,6 @@ static int cmd_open(void *data, const char *input) {
 		case 0:
 		case '?':
 			r_core_cmd_help (core, help_msg_o_);
-			eprintf ("Usage: o-#, o-! or o-*, where # is the filedescriptor number\n");
 		}
 		break;
 	case '.': // "o."
@@ -1303,8 +1389,8 @@ static int cmd_open(void *data, const char *input) {
 			}
                         char *uri = r_str_newf ("malloc://%d", len);
 			ut8 *data = calloc (len, 1);
-			r_core_read_at (core, core->offset, data, len);
-                        RIODesc *fd = r_io_open (core->io, uri, R_IO_READ | R_IO_WRITE, 0);
+			r_io_read_at (core->io, core->offset, data, len);
+                        RIODesc *fd = r_io_open (core->io, uri, R_PERM_R | R_PERM_W, 0);
                         if (fd) {
                                 r_io_desc_write (fd, data, len);
                         }
@@ -1322,7 +1408,7 @@ static int cmd_open(void *data, const char *input) {
 			break;
 		case 'd': // "ood" : reopen in debugger
 			if (input[2] == 'r') { // "oodr"
-				r_core_cmdf ("dor %s", input + 3);
+				r_core_cmdf (core, "dor %s", input + 3);
 				r_core_file_reopen_debug (core, "");
 			} else if ('?' == input[2]) {
 				r_core_cmd_help (core, help_msg_ood);
@@ -1344,7 +1430,7 @@ static int cmd_open(void *data, const char *input) {
 					r_core_cmd_help (core, help_msg_oonn);
 					break;
 				}
-				perms = (input[3] == '+')? R_IO_READ|R_IO_WRITE: 0;
+				perms = (input[3] == '+')? R_PERM_R|R_PERM_W: 0;
 				r_core_file_reopen (core, input + 4, perms, 0);
 				// TODO: Use API instead of !rabin2 -rk
 				if (desc) {
@@ -1355,7 +1441,7 @@ static int cmd_open(void *data, const char *input) {
 				break;
 			}
 
-			perms = ('+' == input[2])? R_IO_READ | R_IO_WRITE: 0;
+			perms = ('+' == input[2])? R_PERM_R | R_PERM_W: 0;
 			r_core_file_reopen (core, input + 3, perms, 0);
 			break;
 		case '+': // "oo+"
@@ -1363,17 +1449,19 @@ static int cmd_open(void *data, const char *input) {
 				r_core_cmd_help (core, help_msg_oo_plus);
 			} else if (core && core->io && core->io->desc) {
 				int fd;
+				int perms = R_PERM_RW;
 				if ((ptr = strrchr (input, ' ')) && ptr[1]) {
 					fd = (int)r_num_math (core->num, ptr + 1);
 				} else {
 					fd = core->io->desc->fd;
+					perms |= core->io->desc->perm;
 				}
-				if (r_io_reopen (core->io, fd, R_IO_READ | R_IO_WRITE, 644)) {
+				if (r_io_reopen (core->io, fd, perms, 644)) {
 					SdbListIter *iter;
 					RIOMap *map;
 					ls_foreach_prev (core->io->maps, iter, map) {
 						if (map->fd == fd) {
-							map->flags |= R_IO_WRITE;
+							map->perm |= R_PERM_WX;
 						}
 					}
 				}
@@ -1388,7 +1476,25 @@ static int cmd_open(void *data, const char *input) {
 				} else {
 					fd = core->io->desc->fd;
 				}
-				r_io_reopen (core->io, fd, R_IO_READ, 644);
+				if (r_config_get_i (core->config, "cfg.debug")) {
+					RBinFile *bf = r_core_bin_cur (core);
+					if (bf && r_file_exists (bf->file)) {
+						char *file = strdup (bf->file);
+						r_core_cmd0 (core, "ob-*");
+						r_io_close_all (core->io);
+						r_config_set (core->config, "cfg.debug", "false");
+						ut64 orig_baddr = sdb_num_get (core->sdb, "orig_baddr", 0);
+						r_bin_set_baddr (core->bin, orig_baddr);
+						r_config_set_i (core->config, "bin.baddr", orig_baddr);
+						r_core_bin_rebase (core, orig_baddr);
+						r_core_cmdf (core, "o %s", file);
+						free (file);
+					} else {
+						eprintf ("Nothing to do.\n");
+					}
+				} else {
+					r_io_reopen (core->io, fd, R_PERM_R, 644);
+				}
 			}
 			break;
 		case '?':
@@ -1403,11 +1509,17 @@ static int cmd_open(void *data, const char *input) {
 				eprintf ("This command is disabled in sandbox mode\n");
 				return 0;
 			}
+			if (core->current_task != core->main_task) {
+				eprintf ("This command can only be executed on the main task!\n");
+				return 0;
+			}
 			// memleak? lose all settings wtf
 			// if load fails does not fallbacks to previous file
+			r_core_task_sync_end (core);
 			r_core_fini (core);
 			r_core_init (core);
-			if (!r_core_file_open (core, input + 2, R_IO_READ, 0)) {
+			r_core_task_sync_begin (core);
+			if (!r_core_file_open (core, input + 2, R_PERM_R, 0)) {
 				eprintf ("Cannot open file\n");
 			}
 			(void)r_core_bin_load (core, NULL, baddr);
